//================================================================
// HW 3(a) Matrix multiply continued -- Moderately sized matrices
// 
// Spectral methods offer accuracy that increases exponentially with
// the number of grid-points.  To reap this benefit, one needs to
// choose the location of grid-points carefully.  Furthermore,
// operations like differentiation become dense (i.e. for a grid with
// N points, one would use all N points go compute derivatives).


#include <cmath>
#include <iostream>
#include <iomanip>



// return the correct location of the grid-points, given a grid with N
// points.  The pointer 'x' must point to an allocated array of length
// N+1
void ChebyshevCollocationPoints(const int N, double* x) {
  for(int k=0; k<=N; ++k)
    x[k] = cos(M_PI *k / N );
};


// differentiation matrix for a Chebyshev-grid with collocation points
// x.  The pointer 'x' must point to an allocated array of length N+1,
// filled with ChebyshevCollocationPoints().  The pointer result must
// point to an allocated array of length (N+1)*(N+1).
//
// result[(N+1)*i+j] = D_{ij} 

//  (code not optimized, because the matrix will be computed only
// once)
void DifferentiationMatrix(const int N, const double* x, double* D) {
  const int n=N+1; // translate from spectral notation to programming notation
  D[0*n+0] =  (2.*N*N+1)/6.;
  D[N*n+N] = -(2.*N*N+1)/6.;
  for(int j=1; j<N; ++j) 
    D[j*n+j] = -x[j]/(2*(1-x[j]*x[j]));

  // off-diagonal elements
  for(int i=0; i<=N; ++i) {
    for(int j=0; j<=N; ++j) {
      if(i!=j) {
	const double sgn = 1-2.*((i+j) % 2);  // (-1)^(i+j) 
	//std::cout << "i+j=" << i+j << ", sgn=" << sgn << std::endl;
	D[i*n+j]= sgn/(x[i]-x[j]); 
	// adjust for edge:
	if(i==0 || i==N) D[i*n+j] *= 2;
	if(j==0 || j==N) D[i*n+j] *= 0.5;
      }
    }
  }
}

//================================================================
// trial function 
//================================================================
double f(const double x) {
  return exp(sin(6*x));
}


double df(const double x) {
  return 6*cos(6*x)*exp(sin(6*x));
}

//================================================================
// Test routine:  Differentiate & compare to analytical answer
//================================================================

void OneD_Test(const int N) {
  const int n=N+1;

  // setup of spectral information
  double* x=new double[n];
  ChebyshevCollocationPoints(N,x);
  double* D=new double[n*n];  
  DifferentiationMatrix(N,x,D);

  // vector to be differentiated
  double* data=new double[n];
  for(int i=0; i<n; ++i) data[i] = f(x[i]);

  // storage for derivative
  double* ddata=new double[n];

  //================
  // Compute Derivative
  //================
  // this, of course, is a very inefficient implementation.
  for(int i=0; i<n; ++i) {
    ddata[i]=0;
    for(int j=0; j<n; ++j) {
      ddata[i] += D[i*n+j]*data[j];
    }
  }
  
  //================
  // Compute error
  //================
  double err=0;
  for(int i=0; i<n; ++i) {
    double tmp=ddata[i] - df(x[i]);
    err += tmp*tmp;
  }
  err=sqrt(err/n);

  //================
  // Output error
  //================

  std::cout << N << "  " << err << std::endl;
}

int main() {
  for(int N=8; N<=64; N+=8)
    OneD_Test(N);
}
